import json
import pickle
import elasticsearch
import psycopg2
from google.cloud.storage import Client

class ElasticsearchClient:
    
    def __init__(self, host, user, password):
        """Instantiate an elasticsearch client.
        
        Parameters
        ----------
        host: str
        user: str
        password: str

        Returns
        -------
        An elasticsearch client.
        """

        self.host = host
        self.user = user
        self.password = password
        
        try:
            es = self._connect()
            if es.ping():
                self._es = es
        except:
            raise ConnectionError(f"Connection to {self.host} failed")
    
    def _connect(self):
        return elasticsearch.Elasticsearch(
            hosts=[self.host],
            http_auth=(self.user, self.password),
            send_get_body_as='POST'
        )
        
    def query(self, index, body):
        return self._es.search(index=index, body=body)
         
class DatabaseClient:

    def __init__(self, host, database, user, password, port=5432):
        """Instantiate a PostgreSQL client.
        
        Parameters
        ----------
        host: str
        database: str
        user: str
        password: str
        port: int

        Returns
        -------
        A PostgreSQL client.
        """
        
        self.host = host
        self.database = database
        self.user = user
        self.password = password
        self.port = port
        
        try:
            conn = self._connect()
            conn.close()
        except:
            raise ConnectionError(f"Connection to {self.host} failed")
    
    def _connect(self):
        return psycopg2.connect(**self.__dict__)

    def query(self, query_):

        conn = None

        try:            
            conn = self._connect()
            cursor = conn.cursor()
            cursor.execute(query_)
            return cursor.fetchall()
        
        except:
            raise
            
        finally:            
            if conn:
                cursor.close()
                conn.close()
    
class GCSClient:
    
    def __init__(self, bucket):
        """Instantiate a Google Cloud Storage client.
        
        Parameters
        ----------
        bucket: str

        Returns
        -------
        A Google Cloud Storage client.
        """
        self.bucket = bucket
        self._client = self._connect()
        
    def _connect(self):
        return Client()
    
    def download(self, blob):
        return self._client.get_bucket(self.bucket).get_blob(blob)

    def upload(self, blob, obj):
        try:
            self._client.get_bucket(self.bucket).blob(blob).upload_from_string(
                data=json.dumps(obj),
                content_type='application/json'
            )
        except TypeError:
            self._client.get_bucket(self.bucket).blob(blob).upload_from_string(
                data=pickle.dumps(obj),
                content_type='application/octet-stream'
            )
        except Exception as e:
            raise e


import os
from pynidus.clients import ElasticsearchClient, DatabaseClient, GCSClient
from pynidus.errors import ErrorLogger
from pynidus.ab import ABTest

class MLTBase:
    
    def __init__(self, **kwargs):
        
        es_config = kwargs.get('es_config')
        pg_config = kwargs.get('pg_config')
        gcs_config = kwargs.get('gcs_config')
        bugsnag_config = kwargs.get('bugsnag_config')
        ab_config = kwargs.get('ab_config')

        if es_config:
            self.es_client = ElasticsearchClient(**es_config)

        if pg_config:
            self.pg_client = DatabaseClient(**pg_config)

        if gcs_config:
            self.gcs_client = GCSClient(**gcs_config)

        if bugsnag_config:
            self.error_logger = ErrorLogger(**bugsnag_config)

        if ab_config:
            self.ab = ABTest(**ab_config)

        





import datetime
import json
import psycopg2
import random
from pynidus.clients import DatabaseClient

class ABTest(DatabaseClient):
    
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @staticmethod
    def heads_or_tails():
        return random.choice(['A', 'B'])[0]

    def _get_or_create_user_tests(self, user_id, ab_test_id):
        
        conn = self._connect()

        cursor = conn.cursor()
        cursor.execute(f"SELECT * FROM ab_tests WHERE user_id='{user_id}' AND ab_test_id='{ab_test_id}'")
        user = cursor.fetchone()

        if not user:
            coin_flip = self.heads_or_tails()
            user = { 
                'user_id': user_id, 
                'created_at': datetime.datetime.now(),
                'ab_test_id': ab_test_id,
                'properties': json.dumps({
                    'ab_test_group': coin_flip
                })
            }
            cursor.execute('INSERT INTO ab_tests (user_id, created_at, ab_test_id, properties) VALUES(%(user_id)s, %(created_at)s, %(ab_test_id)s, %(properties)s)', user)
            conn.commit()
            user['properties'] = json.loads(user['properties'])
            user = list(user.values())
            print(f"User {user_id} has been assigned to group {coin_flip}.")

        cursor.close()
        conn.close()
        
        return user

    def _log_event(self, user_id, event_properties):

        conn = self._connect()

        cursor = conn.cursor()
        cursor.execute('INSERT INTO events (user_id, created_at, properties) VALUES(%(user_id)s, %(created_at)s, %(properties)s)', { 
            'user_id': user_id, 
            'created_at': datetime.datetime.now(),
            'properties': json.dumps(event_properties)
        })
        conn.commit()
        cursor.close()
        conn.close()


import logging
import bugsnag

class ErrorLogger:

    def __init__(self, api_key, release_stage):

        self.api_key = api_key
        self.release_stage = release_stage
            
        bugsnag.configure(
            api_key=self.api_key,
            release_stage=self.release_stage
        )
        
        self.logger = logging.getLogger('bugsnag')
        
        handler = bugsnag.handlers.BugsnagHandler()
        handler.setLevel(logging.ERROR)

        self.logger.addHandler(handler)

    def notify(self, e):
        bugsnag.notify(e)
        self.logger.error(e)
    
class ConfigError(Exception):

    def __init__(self, message):
        super().__init__(message)